﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.Diagnostics;

/*
 *Goal State:
 *1 2 3
 *8 0 4
 *7 6 5
 *
 * odds can move in two directions
 * evens can move in three
 * 0 can move in four
 * 
 * Steps:
 * Check for actives with distance 0
 * Find best active state
 * Find all possible moves from state
 * Add all possible moves to active that aren't already in closed or active
 * Repeat
 */

namespace AI
{
    class Alg
    {
        ///<summary>
        ///This class is the implementation of the A* algorithm
        ///</summary>

        //TODO replace ArrayLists with Lists
        private ArrayList active;
        private ArrayList closed;
        private ArrayList distance;
        private ArrayList depth;
        private ArrayList nilsson;

        private Dist dist;
        private Nil nil;

        public Alg(int[] initial, out string output)
        {
            active = new ArrayList();
            closed = new ArrayList();
            distance = new ArrayList();
            depth = new ArrayList();
            nilsson = new ArrayList();

            dist = new Dist();
            nil = new Nil();

            active.Add(initial);
            depth.Add(1);
            distance.Add(dist.Check(initial));
            nilsson.Add(nil.Check(initial));
            int i = 1;
            while (!distance.Contains(0))
            {
                Debug.WriteLine("Loop: " + i);
                i++;
                Loop();
            }
            output = ConstructOut();
        }

        private string ConstructOut()
        {
            int i = distance.IndexOf(0);
            string s = "Goal Found" + System.Environment.NewLine;
            s += ("Final depth: " + (int)depth[i] + System.Environment.NewLine);
            s += ("Number of explored nodes: " + closed.Count + System.Environment.NewLine);
            s += ("Number of unexplored nodes: " + active.Count + System.Environment.NewLine);
            s += ("Total nodes: " + (closed.Count + active.Count) + System.Environment.NewLine);
            s += "Final state: ";
            foreach (int n in (int[])active[i])
            {
                s += n;
            }
            s += System.Environment.NewLine;
            return s;
        }

        private void Loop()
        {
            int best = FindBest();

            int[] use = (int[])active[best];

            int newDepth = (int)depth[best];
            newDepth++;

            closed.Add(use);
            active.RemoveAt(best);
            distance.RemoveAt(best);
            depth.RemoveAt(best);
            nilsson.RemoveAt(best);

            int open = FindOpen(use);

            /*
             * odds can move in two directions
             * evens can move in three
             * 0 can move in four
             */

            if (open == 0)
            {
                //0 is centered
                ShiftFour(use, newDepth);
            }
            else if(open % 2 == 1)
            {
                //0 is in an odd space
                ShiftTwo(open, use, newDepth);
            }
            else
            {
                //0 is in an even space
                ShiftTwo(open, use, newDepth);
                ShiftThree(open, use, newDepth);
            }
        }

        private int FindBest()
        {
            int best = 0;
            int bestV = 5000;
            for (int n = 0; n < active.Count; n++)
            {
                if ((int)distance[n] + (int)depth[n] + (3*(int)nilsson[n]) < bestV)
                {
                    bestV = (int)distance[n] + (int)depth[n] + (3*(int)nilsson[n]);
                    best = n;
                }
            }

            return best;
        }

        private int FindOpen(int[] use)
        {
            int open = 0;

            for (int i = 0; i < 9; i++)
            {
                if (use[i] == 0)
                {
                    open = i;
                    break;
                }
            }

            return open;
        }

        private void TryAdd(int[] i, int inDepth)
        {
            bool fail = false;

            for (int x = 0; x < active.Count; x++)
            {
                if (!fail)
                {
                    int[] a = (int[])active[x];
                    int count = 0;
                    for (int n = 0; n < 9; n++)
                    {
                        if (i[n] != a[n])
                        {
                            break;
                        }
                        else
                        {
                            count++;
                            if (count == 9)
                            {
                                fail = true;
                            }
                        }
                    }
                }
            }

            if (!fail)
            {
                for (int x = 0; x < closed.Count; x++)
                {
                    if (!fail)
                    {
                        int[] a = (int[])closed[x];
                        int count = 0;
                        for (int n = 0; n < 9; n++)
                        {
                            if (i[n] != a[n])
                            {
                                break;
                            }
                            else
                            {
                                count++;
                                if (count == 9)
                                {
                                    fail = true;
                                }
                            }
                        }
                    }
                }
            }

            if (!fail)
            {
                active.Add(i);
                depth.Add(inDepth);
                distance.Add(dist.Check(i));
                nilsson.Add(nil.Check(i));
            }
        }

        /*
         *Goal State:
         *1 2 3
         *8 0 4
         *7 6 5
         */

        private void ShiftTwo(int open, int[] use, int newDepth)
        {
            int a;
            int b;

            int[] newA = (int[])use.Clone();
            int[] newB = (int[])use.Clone();

            switch (open)
            {
                case 8:
                    a = 1;
                    b = open - 1;
                    break;
                case 1:
                    a = open + 1;
                    b = 8;
                    break;
                default:
                    a = open + 1;
                    b = open - 1;
                    break;
            }
            newA[open] = newA[a];
            newA[a] = 0;
            TryAdd(newA, newDepth);            

            newB[open] = newB[b];
            newB[b] = 0;
            TryAdd(newB, newDepth);                  
        }

        /*
         *Goal State:
         *1 2 3
         *8 0 4
         *7 6 5
         */

        private void ShiftThree(int open, int[] use, int newDepth)
        {
            int[] newA = (int[])use.Clone();
            newA[open] = newA[0];
            newA[0] = 0;
            TryAdd(newA, newDepth); 
        }

        private void ShiftFour(int[] use, int newDepth)
        {
            for (int i = 1; i < 5; i++)
            {
                int[] a = (int[])use.Clone();
                a[0] = a[i * 2];
                a[i * 2] = 0;
                TryAdd(a, newDepth);
            }
        }
    }
}
